;; SaveEventsAsMIDI.lisp
;; written by Paul Pelton

;; slight fix and other modifications by Soren Goodman

(in-package :cl-user)


(defconstant kMThdHeaderBytes '(#x4D #x54 #x68 #x64))
(defconstant kMThdHeaderLenBytes '(0 0 0 6))
(defconstant kMThdHeaderFormatBytes '(0 1))
(defconstant kMTrkHeaderBytes '(#x4D #x54 #x72 #x6B))
(defconstant kMTrkEndBytes '(#x00 #xFF #x2F #x00))
(defconstant kPPQNValue #x30)

(defun save-as-midi (outfilename events &optional (author "Soren Goodman"))
   (let ((tracks (create-midi-tracks author events)))
      (with-open-stream (file (create-midi-file outfilename))
        (push (create-MThd (length tracks)) tracks)
        (write-to-midi-file file (setq testtracks tracks))) ;; now includes MThd header
     ;(close (open outfilename)) ;;;This was needed in mcl I think...
     outfilename)) ;; close and return outfilename
   
(defun create-midi-file (outfilename)
  (open outfilename :direction :output :if-exists :supersede
        :if-does-not-exist :create :element-type 'unsigned-byte))

;;changed :IF-EXISTS from :ERROR to :SUPERSEDE

(defun write-to-midi-file (file listOfChunks)
   (if (null listOfChunks) ()
      (progn
       (dolist (byte (first listOfChunks))
         (write-byte byte file))
       (write-to-midi-file file (rest listOfChunks)))))

(defun create-midi-tracks (author copeEvents)
  (let ((tracks (create-tempo-track)))
    (dotimes (channel 16 (reverse tracks))
      (let ((channelEvents (get-channel-events (1+ channel) copeEvents)))
        (if channelEvents
          (push (create-MTrk channelEvents) tracks))))))

(defun create-tempo-track ()
   (list (append kMTrkHeaderBytes (split-bytes (length kMTrkEndBytes) 4) kMTrkEndBytes)))

;; !!!! this should be in ACLMidi.Lisp
(defun get-channel-events (channel events)
   (cond
    ((null events) ())
    ((= channel (fourth (first events)))
     (cons (first events) (get-channel-events channel (rest events))))
    (t (get-channel-events channel (rest events)))))

;; this is adapted to Lisp from the C code at
;; www.borg.com/~jglatt/tech/midifile.htm
(defun make-var-len (value)
   (let ((buffer (list (logand #x7F value))))
      (loop while (not (zerop (setq value (ash value -7))))
         do (push (logior (logior #x80 (logand #x7F value))) buffer))
 buffer))

(defun create-MThd (numtracks)
   ;; assume that 1 <= numtracks <= 16
   (append
    kMThdHeaderBytes
    kMThdHeaderLenBytes
    kMThdHeaderFormatBytes
    (list 0 numtracks 0 kPPQNValue)))

(defun create-MTrk (events)
   (if (null events) ()
      ;; assume 2 bytes for length of track data
      (let* ((mtrkData
              (append
               (create-midi-track-data
                (fix-deltatime 0 (sort-by-deltatime (create-midi-events events))))
               kMTrkEndBytes)))
         ;; mtrkData now includes kMTrkEndBytes
         (append kMTrkHeaderBytes (split-bytes (length mtrkData) 4) mtrkData))))

;; split a long integer into its high byte and low byte
(defun split-bytes (num count)
   (let ((bytes ()))
      (dotimes (i count bytes)
         (push (get-byte i num) bytes))))

;; byteIndex starts at 0 (rightmost) and goes to whatever (leftmost)
(defun get-byte (byteIndex num)
   (ldb (byte 8 (* 8 byteIndex)) num))

;; midiEvents are ((deltaTime (byte3 byte2 byte1))(deltaTime (byte3 byte2 byte1))...),
;; are sorted in the order they should be in the file and their deltaTimes are
;; relative to each other (each relative to the previous).
(defun create-midi-track-data (midiEvents)
   (if (null midiEvents) ()
      (let ((midiEvt (first midiEvents)))
        (append
         (make-var-len (first midiEvt))
         (second midiEvt)
         (create-midi-track-data (rest midiEvents))))))

(defun create-midi-events (copeEvents)
   (if (null copeEvents) ()
      (let* ((event (first copeEvents))
             (ontime (first event)))
         (append
          (cond ((= (fifth event) 255)
                 (list (list (convert-ontime-to-deltatime ontime) (make-midi-pc-msg event))))
                (t
                 (list
                  (list
                   (convert-ontime-to-deltatime ontime) (make-midi-note-msg event #x90))
                  (list
                   ;;(convert-ontime-to-deltatime (+ ontime (third event)) ...
                   (convert-ontime-to-deltatime (third event)) (make-midi-note-msg event #x80)))))
          (create-midi-events (rest copeEvents))))))

;; #x90 = note-on
;; #x80 = note-off
(defun make-midi-note-msg (copeEvent flag)
   (list (logior (1- (fourth copeEvent)) flag) (second copeEvent) (fifth copeEvent)))

;; #xC0 = program change
(defun make-midi-pc-msg (copeEvent)
   (list (logior (1- (fourth copeEvent)) #xC0) (second copeEvent)))
                                               
(defun sort-by-deltatime (midiEvents)
   (sort midiEvents #'< :key #'car))


(defun fix-deltatime (lasttime midiEvents)
   (if (null midiEvents) ()
      (let* ((midiEvt (first midiEvents))
            (newLastTime (first midiEvt)))
         (cons
          (list (- newLastTime lasttime) (second midiEvt))
          (fix-deltatime newLastTime (rest midiEvents))))))

(defun convert-ontime-to-deltatime (copeOntime)
   (round (* (/ kPPQNValue 1000) copeOnTime)))


;;not in use
(defun strip-unknowns (copeEvents)
  (let ((event (first copeEvents)))
    (cond
     ((null copeEvents) ())
     ((> (fifth event) 127) (strip-unknowns (rest copeEvents)))
     (t (cons event (strip-unknowns (rest copeEvents)))))))

