;;Copyright (C) 2002 Soren Goodman
;;Textures Music Score Generator
;;See readme.txt for more info

;;;saveascsound.lisp
;;;march 2002
;;;textures thesis project


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;note!  These must be made lowercase!!!;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *csound-ftables*
  (list '(sine-table 1)
        '(square-table 2)))


(defun save-csound (filename data)
  (save-as-csound filename (perform data csound-score) *csound-ftables*)
  (format t "~%~S" filename))

(defun create-save-csound (filename symbol)
  (if (listp symbol) 
      (let ((function (first symbol))
            (num-attacks (get-data (second symbol))))
        (create function num-attacks)
          (save-csound filename function))
    (progn (create symbol)
      (save-csound filename symbol))))

(defun create-n-csave (filename symbol)
  (let ((file (concatenate 'string "c:\\csound\\textures\\" filename ".sco")))
    (create-save-csound file symbol)))


;;this value is for converting midi dynamics to csound.  A dynamic value of 127=this value.
;;if you are getting clipping, lower this value.
(defparameter *volume-scaling-for-csound* 5000) 

;;there is a possibility of having a more interesting conversion, thus this is a seperate function
(defun midi-volume-to-csound (midi-volume)
  (float (* *volume-scaling-for-csound* (/ midi-volume 127))))


(defmethod perform ((input phrase) (output csound-score) &optional (clock 0))
  (declare (ignore clock))
  (setf *perform-error* 'perform-phrase)
  (when (fboundp 'gc)(gc)) ;;GC - garbage collector - is not common lisp
  (setf (:data output)
    (let* ((instrument (get-data (:instrument input)))
           (instrument (if instrument instrument
                         default-csound-instrument))
           (clock-offset (round (get-data (:clock-offset input))))
           (tempo (get-data (:tempo input)))
           (tempo (if (null tempo) *tempo* tempo))
           (rhythm (durations-to-rhythm (check-list (:rhythm input))))
           (pitch (check-list (:pitch input)))
           (pitch-template (:pitch-template instrument)) ;;mostly for percussion
           (pitch (make-pch (if pitch-template 
                                (convert-pitches-to-template pitch pitch-template)
                              pitch)))
           (duration (check-list (:duration input)))
           (duration (if duration duration
                       '(1)))
           (dynamics (check-list (:dynamics input)))
           (dynamics (if dynamics dynamics
                       '(127)))
           (extra (check-list (:extra input)))
           (instrument-nums (check-list (:instrument-nums input)))
           (num-attacks (num-attacks rhythm))
           (pitch (check-and-modify-length pitch num-attacks))
           (duration (check-and-modify-length duration num-attacks)) 
           (dynamics (check-and-modify-length dynamics num-attacks))
           (start-measure (beats-to-measures clock-offset))
           (end-measure (beats-to-measures (last-beat input t))))
      (if (or (null pitch) (null rhythm)) 
          (format t "nil")
        (progn (format t "at clock ~S from mm.~S to mm.~S."  clock-offset start-measure end-measure)
          (make-phrase-csound instrument-nums rhythm pitch duration dynamics extra tempo clock-offset))))))


(defun make-phrase-csound (instrument-nums rhythm pitch durations dynamics extra tempo &optional (clock 0)
                                      &aux (clock (round clock)))
 (setq *perform-error* 'make-phrase-csound)
  (do* ((clock clock (1+ clock))
        output)
       ((null rhythm) (reverse output))
    (if (zerop (first rhythm))
        (setq rhythm (rest rhythm))
      (let* ((time-on (float (* clock (/ 60 tempo))))
             (duration (- (float (+ time-on
                                   (if (zerop (first duration))
                                       (/ *staccato* 1000) ;;csound wants seconds, not ms
                                     (* (/ 60 tempo)
                                        (first durations)))))
                          time-on))
             (dynamic (midi-volume-to-csound (first dynamics)))
             (event 
              (append (list (first instrument-nums) 
                            time-on 
                            duration
                            (first pitch)
                            dynamic)
                      (mapcar #'first extra))))
        (push event output)
        (setq rhythm (rest rhythm)
            pitch (rest pitch)
            durations (rest durations)
            dynamics (rest dynamics)
            instrument-nums (rest instrument-nums)
            extra (mapcar #'rest extra))))))
             
                    
                    
             

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;csound gen-table functions;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;        



(defparameter *author* "Soren Goodman")

(defparameter *csound-header*
  (format nil 
      ";;;This file was created by ~A on ~a
;;;using the Textures Music Score Generator.
;;;See www.theory.org/~~soren/textures for more information."  *author* (date-string)))

(defparameter *default-ftable-size* (expt 2 16))

;;size of 0 means get size from file, but can only be read by the loscil opcode, which can't
;;read partway into a file or read a file in reverse.  GEN01
(defun sample-table (fnum file &key (size 0) (skip 0))
  (let ((file (concatenate 'string "\"" file "\""))) ;;this is necessary to print the " quotes
    (list (concat-symbol "f" fnum) 0 size 1 file 0 0 0)))

;;uses gen10 to create harmonics with harmonic partials of defined strengths 0-1.
(defun even-harmonic-table (fnum &key (size *default-ftable-size*) (harmonics '(1)))
  (append (list (concat-symbol "f" fnum) 0 size 10) harmonics))

;;GEN10 with only a single sin fundamental
(defun sine-table (fnum &key (size *default-ftable-size*))
  (even-harmonic-table fnum :size size :harmonics '(1)))

;;GEN27 with coordinates to create a simple square table.
(defun square-table (fnum &key (size *default-ftable-size*))
  (linear-segments-table fnum :size size :segments '(0 1 .5 1 .5 -1 1 -1)))

;;this uses gen27 to create a linear segments table.  The segments is in the format
;;(time1% value1 time2% value2 ...).  In other words, time% is scaled to size of the table, and value
;;is the level that should be reached by that time.  If two times are identical, the second will have a 
;;single sample added.  The default value of SEGMENTS describes a square wave.
(defun linear-segments-table (fnum &key (scale-to-1 t)(size *default-ftable-size*) (segments '(0 1 .5 1 .5 -1 1 -1)))
  (let* ((times (do ((collect)
                     (n 0 (+ n 2)))
                    ((>= n (length segments)) (reverse collect))
                     (push (nth n segments) collect)))
         (values (do ((collect)
                      (n 1 (+ n 2)))
                     ((>= n (length segments)) (reverse collect))
                   (push (nth n segments) collect)))
         (times (mapcar #'floor (scale-to-n times size)))
         (times (add1-to-duplicates (sort times #'<)))
         (segments (mapcan #'list times values))
         (gen-num (if scale-to-1 27 -27)))    
    (append (list (concat-symbol "f" fnum) 0 size gen-num) segments)))





;;gen21
;;type options are uniform, linear, triangular, exponential, biexponential, gaussian, cauchy, positive cauchy and poisson.
(defun random-table (fnum &key (size *default-ftable-size*) (type 'uniform)(max 1))
  (let ((type (cond ((eql type 'uniform) 1)
                    ((eql type 'linear) 2)
                    ((eql type 'triangular) 3)
                    ((eql type 'exponential) 4)
                    ((eql type 'biexponential) 5)
                    ((eql type 'gaussian) 6)
                    ((eql type 'caushy) 7)
                    ((eql type 'positive-caushy) 8)
                    ((eql type 'poisson) 11))))
    (list (concat-symbol "f" fnum) 0 size 21 type max)))

;;;;;;;;;;;;;;;;;;
;;;Save-As-Csound;;;
;;;;;;;;;;;;;;;;;;

#|This function saves a list of lists as a textfile with a #/newline between each list.
In addition, there is an optional parameter that contains a list of ftables.  The ftable functions
(for sine, square, random, etc) are contained above.  If there are any ftables present, they are inserted
at the front of the document.  

The input format is (saveascsound filename eventslist &optional ftables)

The events input is expected to be an ordered list of pfields.  I.e. ((p1 p2 p3 p4 ...pn)(p1 p2 p3 p4 ... pn))
where p1 is the instrument number, p2 is the time on in ms, p3 is the duration is ms, and the rest are variable in csound,
so there will be a way to modify the order, but in here will likely be p4=pch, p5=amp (0 - 32000) and the rest are
instrument dependent.
|#



(defun save-as-csound (filename events &optional (ftables *csound-ftables*) (header *csound-header*))
  (with-open-stream (file (open filename :direction :output :if-exists :supersede
                                :if-does-not-exist :create :element-type 'character))
    (when header 
      (write-sequence header file)
      (write-char #\newline file)
      (write-char #\newline file))
    (when ftables 
      (setq ftables (mapcar #'get-data ftables))
      (write-sequence ";;;Function tables." file)
      (write-char #\newline file)
      (if (listp (first ftables))
                      (mapcar (lambda (table) 
                                (write-list table file 1 #\space)
                                (write-char #\newline file)) 
                        ftables)
                    (write-list ftables file 1 #\space))
      
      (write-char #\newline file)
      (write-char #\newline file))
    
    (when events 
      (write-sequence ";;;Instrument events." file)
      (write-char #\newline file)
      (write-sequence ";instr#(p1)" file)
      (write-char #\tab file)
      (write-sequence "start(p2)" file)
      (write-char #\tab file)
      (write-char #\tab file)
      (write-sequence "dur(p3)" file)
      (write-char #\tab file)
      (write-char #\tab file)
      (write-sequence "pitch(p4)" file)
      (write-char #\tab file)
      (write-char #\tab file)
      (write-sequence "volume(p5)" file)
      (write-char #\newline file)
      (mapcar (lambda (event) (write-char #\i file)
                           (write-list event file)(write-char #\newline file)) events))
    (close file)
    ))

;;writes each item of list, without ( and ), seperated by SEPERATOR into FILE.
(defun write-list (seq file &optional (num-seperators 2)(seperator #\tab))
  (let ((string (princ-to-string seq)))
    (loop for n from 1 below (length string) do
          (let ((char (aref string n)))
            (cond ((char= char #\space)
                   (loop for n from 1 to num-seperators do
                         (write-char seperator file)))
                  ((or (char= char #\) )
                       (char= char #\( )))
                  (t    ;;need to make sure f and i are lowercase or csound will shit.
                   (write-char (char-downcase char) file)))))))






